/*
 *
 * Copyright (c) 2021 Texas Instruments Incorporated
 *
 * All rights reserved not granted herein.
 *
 * Limited License.
 *
 * Texas Instruments Incorporated grants a world-wide, royalty-free, non-exclusive
 * license under copyrights and patents it now or hereafter owns or controls to make,
 * have made, use, import, offer to sell and sell ("Utilize") this software subject to the
 * terms herein.  With respect to the foregoing patent license, such license is granted
 * solely to the extent that any such patent is necessary to Utilize the software alone.
 * The patent license shall not apply to any combinations which include this software,
 * other than combinations with devices manufactured by or for TI ("TI Devices").
 * No hardware patent is licensed hereunder.
 *
 * Redistributions must preserve existing copyright notices and reproduce this license
 * (including the above copyright notice and the disclaimer and (if applicable) source
 * code license limitations below) in the documentation and/or other materials provided
 * with the distribution
 *
 * Redistribution and use in binary form, without modification, are permitted provided
 * that the following conditions are met:
 *
 * *       No reverse engineering, decompilation, or disassembly of this software is
 * permitted with respect to any software provided in binary form.
 *
 * *       any redistribution and use are licensed by TI for use only with TI Devices.
 *
 * *       Nothing shall obligate TI to provide you with source code for the software
 * licensed and provided to you in object code.
 *
 * If software source code is provided to you, modification and redistribution of the
 * source code are permitted provided that the following conditions are met:
 *
 * *       any redistribution and use of the source code, including any resulting derivative
 * works, are licensed by TI for use only with TI Devices.
 *
 * *       any redistribution and use of any object code compiled from the source code
 * and any resulting derivative works, are licensed by TI for use only with TI Devices.
 *
 * Neither the name of Texas Instruments Incorporated nor the names of its suppliers
 *
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * DISCLAIMER.
 *
 * THIS SOFTWARE IS PROVIDED BY TI AND TI'S LICENSORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TI AND TI'S LICENSORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include "app_common.h"
#include "app_capture_module.h"
#include "app_sensor_module.h"
#include <utils/iss/include/app_iss.h>
#include <utils/perf_stats/include/app_perf_stats.h>
#include <TI/tivx_event.h>
#include <TI/tivx_task.h>

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define MAX_CAPTURE_OBJECT              (2u)

#define APP_BUFFER_Q_DEPTH              (4u)

#define TIVX_TARGET_DEFAULT_STACK_SIZE      (256U * 1024U)
#define TIVX_TARGET_DEFAULT_TASK_PRIORITY1   (8u)


/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

typedef struct {

    SensorObj           sensorObj;
    CaptureObj          captureObj[MAX_CAPTURE_OBJECT];

    vx_char             output_file_path[APP_MAX_FILE_PATH];

    /* OpenVX references */
    vx_context          context;

    tivx_task           task;

    app_perf_point_t    total_perf;

    tivx_event          eventHandle_Finished[APP_MAX_FILE_PATH];
    tivx_event          eventHandle_VerifyDone[APP_MAX_FILE_PATH];
    tivx_event          eventHandle_SensorEnabled[APP_MAX_FILE_PATH];
    tivx_task_create_params_t   taskParams_capture[MAX_CAPTURE_OBJECT];
    tivx_task                   taskHandle_capture[MAX_CAPTURE_OBJECT];
} AppObj;


/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static void app_default_param_set(AppObj *obj);
static vx_status app_init(AppObj *obj);
static void app_deinit(AppObj *obj);
static void add_graph_parameter_by_node_index(vx_graph graph, vx_node node, vx_uint32 node_parameter_index);
static void VX_CALLBACK tivxTask_capture(void *app_var);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

AppObj gAppObj;


/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

vx_int32 app_capture_main(vx_int32 argc, vx_char* argv[])
{
    vx_status                   status = VX_SUCCESS;
    uint32_t                    cnt;
    tivx_task_create_params_t  *capt_prms;
    uint32_t                    isEnabled[MAX_CAPTURE_OBJECT] = {1, 1};

    AppObj *obj = &gAppObj;

    /*Optional parameter setting*/
    app_default_param_set(obj);

    /* Querry sensor parameters */
    status = app_querry_sensor(&obj->sensorObj);

    if (VX_SUCCESS == status)
    {
        status = app_init(obj);

        if (VX_SUCCESS == status)
        {
            for (cnt = 0u; cnt < MAX_CAPTURE_OBJECT; cnt ++)
            {
                capt_prms = &obj->taskParams_capture[cnt];

                if (!isEnabled[cnt])
                    continue;

                status = tivxEventCreate(&obj->eventHandle_VerifyDone[cnt]);
                status |= tivxEventCreate(&obj->eventHandle_Finished[cnt]);
                status |= tivxEventCreate(&obj->eventHandle_SensorEnabled[cnt]);

                tivxTaskSetDefaultCreateParams(capt_prms);

                capt_prms->task_main        = &tivxTask_capture;
                capt_prms->app_var          = cnt;
                capt_prms->stack_ptr        = NULL;
                capt_prms->stack_size       = TIVX_TARGET_DEFAULT_STACK_SIZE;
                capt_prms->core_affinity    = TIVX_TASK_AFFINITY_ANY;
                capt_prms->priority         =
                    TIVX_TARGET_DEFAULT_TASK_PRIORITY1;

                status |= tivxTaskCreate(&obj->taskHandle_capture[cnt],
                    capt_prms);

                if (VX_SUCCESS != status)
                {
                    break;
                }
            }
        }

        if (VX_SUCCESS == status)
        {
            /* Wait for verify to complete for all graphs */
            for (cnt = 0u; cnt < MAX_CAPTURE_OBJECT; cnt ++)
            {
                if (!isEnabled[cnt])
                    continue;

                tivxEventWait(gAppObj.eventHandle_VerifyDone[cnt],
                    TIVX_EVENT_TIMEOUT_WAIT_FOREVER);
            }
        }

        /* Now enable the sensor, this is required only for EVM,
           where this application is tested for IMX390 input.
           For Customer application, individual sensor should be
           enabled in the task, after verify graph is called */
        status = appStartImageSensor(gAppObj.sensorObj.sensor_name,
            ((1 << MAX_CAPTURE_OBJECT) - 1));

        if (VX_SUCCESS == status)
        {
            /* Now unblock the graphs, to start their processing */
            for (cnt = 0u; cnt < MAX_CAPTURE_OBJECT; cnt ++)
            {
                if (!isEnabled[cnt])
                    continue;

                tivxEventPost(gAppObj.eventHandle_SensorEnabled[cnt]);
            }
        }

        if (VX_SUCCESS == status)
        {
            /* Wait for all tasks to finish */
            for (cnt = 0u; cnt < MAX_CAPTURE_OBJECT; cnt ++)
            {
                if (!isEnabled[cnt])
                    continue;

                tivxEventWait(gAppObj.eventHandle_Finished[cnt],
                    TIVX_EVENT_TIMEOUT_WAIT_FOREVER);
            }
        }

        if (VX_SUCCESS == status)
        {
            /* Wait for all tasks to finish */
            for (cnt = 0u; cnt < MAX_CAPTURE_OBJECT; cnt ++)
            {
                if (!isEnabled[cnt])
                    continue;

                tivxEventDelete(&obj->eventHandle_VerifyDone[cnt]);
                tivxEventDelete(&obj->eventHandle_Finished[cnt]);
                tivxEventDelete(&obj->eventHandle_SensorEnabled[cnt]);
                tivxTaskDelete(&obj->taskHandle_capture[cnt]);
            }
        }

        app_deinit(obj);
    }

    return 0;
}


/* ========================================================================== */
/*                          Local Function Definitions                        */
/* ========================================================================== */

static void VX_CALLBACK tivxTask_capture(void *app_var)
{
    vx_graph            graph;
    CaptureObj         *capt_obj;
    uint32_t            graph_created = 0;
    uint32_t            graph_parameter_index = 0;
    vx_status           status;
    uint32_t            cnt;
    vx_graph_parameter_queue_params_t graph_parameterslist[1];

    uint32_t capt_idx = (uint32_t)app_var;

    capt_obj = &gAppObj.captureObj[capt_idx];

    graph = vxCreateGraph(gAppObj.context);
    status = vxGetStatus((vx_reference)graph);
    if (VX_SUCCESS == status)
    {
        graph_created = 1;
        status = vxSetReferenceName((vx_reference)graph, "multi_capture_node");
    }

    if (VX_SUCCESS == status)
    {
        status = app_create_graph_capture(graph, capt_obj, capt_idx);
    }

    if (VX_SUCCESS == status)
    {
        /* input @ node index 0, becomes csirx_graph parameter 1 */
        add_graph_parameter_by_node_index(graph, capt_obj->node, 1);

        graph_parameter_index = 0;

        /* set csirx_graph schedule config such that csirx_graph parameter @ index 0 and 1 are enqueuable */
        graph_parameterslist[0].graph_parameter_index = 0;
        graph_parameterslist[0].refs_list_size = APP_BUFFER_Q_DEPTH;
        graph_parameterslist[0].refs_list = (vx_reference*)&capt_obj->raw_image_arr[0];
        graph_parameter_index++;

        /* Schedule mode auto is used, here we dont need to call vxScheduleGraph
        * Graph gets scheduled automatically as refs are enqueued to it
        */
        status = vxSetGraphScheduleConfig(graph, VX_GRAPH_SCHEDULE_MODE_QUEUE_AUTO,
            graph_parameter_index, graph_parameterslist);
    }

    if (VX_SUCCESS == status)
    {
        //status = tivxSetGraphPipelineDepth(graph, APP_BUFFER_Q_DEPTH);
    }
    if (VX_SUCCESS == status)
    {
        status = vxVerifyGraph(graph);
    }

    if (VX_SUCCESS == status)
    {
        /* enqueue capture buffers for pipeup but dont trigger graph executions */
        for(cnt = 0; cnt < (APP_BUFFER_Q_DEPTH - 2); cnt ++)
        {
            tivxGraphParameterEnqueueReadyRef(
                graph, 0u,
                (vx_reference*)&capt_obj->raw_image_arr[cnt],
                1u, TIVX_GRAPH_PARAMETER_ENQUEUE_FLAG_PIPEUP);
        }

        /* Now enqueue a buffer to trigger csirx_graph scheduling */
        vxGraphParameterEnqueueReadyRef(graph, 0,
            (vx_reference*)&capt_obj->raw_image_arr[(APP_BUFFER_Q_DEPTH - 2)], 1);

        vxGraphParameterEnqueueReadyRef(graph, 0,
            (vx_reference*)&capt_obj->raw_image_arr[(APP_BUFFER_Q_DEPTH - 1)], 1);
    }

    if (VX_SUCCESS == status)
    {
        tivxEventPost(gAppObj.eventHandle_VerifyDone[capt_idx]);
    }

    if (VX_SUCCESS == status)
    {
        tivxEventWait(gAppObj.eventHandle_SensorEnabled[capt_idx],
            TIVX_EVENT_TIMEOUT_WAIT_FOREVER);
    }

    if (VX_SUCCESS == status)
    {
        for (cnt = 0; cnt < 1000; cnt ++)
        {
            uint32_t num_refs;
            vx_object_array captured_frames = NULL;

            /* Get captured frame reference, waits until a reference is available */
            vxGraphParameterDequeueDoneRef(graph, 0,
                (vx_reference*)&captured_frames, 1, &num_refs);

            vxGraphParameterEnqueueReadyRef(graph, 0,
                (vx_reference*)&captured_frames, 1);
        }
    }

    if (VX_SUCCESS == status)
    {
        /* ensure all csirx_graph processing is complete */
        vxWaitGraph(graph);
    }

    if (VX_SUCCESS == status)
    {
        app_delete_capture(capt_obj);
    }

    if (graph_created)
    {
        vxReleaseGraph(&graph);
    }

    /*Signal the completion of csirx graph processing*/
    tivxEventPost(gAppObj.eventHandle_Finished[capt_idx]);
}


static void app_default_param_set(AppObj *obj)
{
    strcpy(obj->sensorObj.sensor_name, SENSOR_SONY_IMX390_UB953_D3);

    obj->sensorObj.num_sensors_found            = 0;
    obj->sensorObj.sensor_features_enabled      = 0;
    obj->sensorObj.sensor_features_supported    = 0;
    obj->sensorObj.sensor_dcc_enabled           = 0;
    obj->sensorObj.sensor_wdr_enabled           = 0;
    obj->sensorObj.sensor_exp_control_enabled   = 0;
    obj->sensorObj.sensor_gain_control_enabled  = 0;

    obj->sensorObj.num_cameras_enabled          = MAX_CAPTURE_OBJECT;
}

static vx_status app_init(AppObj *obj)
{
    vx_status                           status = VX_SUCCESS;
    uint32_t                            cnt;
    uint32_t                            vcId;

    /* Create OpenVx Context */
    obj->context = vxCreateContext();
    status = vxGetStatus((vx_reference)obj->context);

    if (status == VX_SUCCESS)
    {
        tivxHwaLoadKernels(obj->context);
    }

    /* Initialize modules */
    if (status == VX_SUCCESS)
    {
        status = app_init_sensor(&obj->sensorObj, "sensor_obj");
    }

    if (status == VX_SUCCESS)
    {
        for (cnt = 0u; cnt < MAX_CAPTURE_OBJECT; cnt ++)
        {
            vcId = cnt;

            /* Change here RAW_Params based on the sensor.
               Keeping it here same because on EVM, all sensors are same. */
            status = app_init_capture(obj->context,
                &obj->captureObj[cnt],
                &obj->sensorObj.sensorParams.sensorInfo.raw_params, 1u,
                &vcId, 0, "capture_obj", APP_BUFFER_Q_DEPTH);

            if (status != VX_SUCCESS)
            {
                break;
            }
        }
    }

    if (status == VX_SUCCESS)
    {
        appPerfPointSetName(&obj->total_perf , "TOTAL");
    }
    return status;
}

static void app_deinit(AppObj *obj)
{
    uint32_t cnt;

    app_deinit_sensor(&obj->sensorObj);

    for (cnt = 0u; cnt < MAX_CAPTURE_OBJECT; cnt ++)
    {
        app_deinit_capture(&obj->captureObj[cnt], APP_BUFFER_Q_DEPTH);
    }

    tivxHwaUnLoadKernels(obj->context);

    vxReleaseContext(&obj->context);
}

/*
 * Utility API used to add a graph parameter from a node, node parameter index
 */
static void add_graph_parameter_by_node_index(vx_graph graph, vx_node node, vx_uint32 node_parameter_index)
{
    vx_parameter parameter = vxGetParameterByIndex(node, node_parameter_index);

    vxAddParameterToGraph(graph, parameter);
    vxReleaseParameter(&parameter);
}
